module ExoBasic
  class Timer
    attr_reader :name, :start, :prev, :now, :samples

    EPOCH           = Time.utc(1970, 1, 1)
    END_OF_ALL_DAYS = Time.utc(3000, 1, 1)

    def initialize(name=nil)
      @name    = name
      @start   = Time.now
      @prev    = @start
      @now     = @start
      @samples = 0
    end

    def sample
      t         = Time.now
      @prev     = @now
      @now      = t
      @samples += 1
    end

    def tick
      @now = Time.now
    end

    def warp
      @start   = @prev
      @samples = 1
    end

    def delta
      Time.now - @now
    end

    def lifetime
      Time.now - @start
    end

    def diff_from(that)
      @now - that.now
    end

    def merge_with(that)
      if that.start < @start
        @start    = that.start
        @samples += 1
      end
      if that.now > @now
        @now      = that.now
        @samples += 1
      end
    end

    def show(msg=nil)
      "-- TIME_MEASUREMENT #{@name}::#{msg}::#{@samples}\tfrom_prev = #{@now - @prev}\tfrom_start = #{@now - @start}"
    end

    def checkpoint(msg=nil)
      self.sample
      STDERR.puts self.show(msg)
    end

    def self.time_now
      Time.now
    end

    def self.get(date_time)
      if date_time.nil?
        Timer.time_now
      elsif date_time.is_a?(String)
        Time.parse(date_time)
      else
        date_time
      end
    end

    def self.duration_in_days(from, to)
      ((to - from) / 86400).to_i
    end

    def self.duration_in_sec(from, to)
      to - from
    end

    def self.add_days(t, days)
      t + days * 86400
    end

    def self.add_sec(t, sec)
      t + sec
    end

    # sec
    def self.from_epoch(epoch)
      Time.at(epoch)
    end

    # sec
    def self.to_epoch(t)
      t.strftime('%s').to_i
    end

    def self.to_iso8601(t)
      t.iso8601
    end

    def self.year(t)
      t.year
    end

    # [YYYY, (1..366)]
    def self.date_by_day(t)
      [t.year, t.strftime('%-j').to_i]
    end

    def self.date_by_day_to_epoch(tuple)
      Timer.to_epoch(Time.strptime("#{tuple[0]} #{tuple[1]}", '%Y %j'))
    end

    def self.day_of_year(t)
      Timer.date_by_day(t)[1]
    end

    # [YYYY, (1..54), (1..7)]
    # week starts with Monday
    def self.date_by_week(t)
      [t.year, t.strftime('%W').to_i + 1, t.strftime('%u').to_i]
    end

    def self.date_by_week_to_epoch(tuple)
      Timer.to_epoch(Time.strptime("#{tuple[0]} #{tuple[1] - 1} #{tuple[2]}", '%Y %W %u'))
    end

    def self.week_of_year(t)
      Timer.date_by_week(t)[1]
    end

    def self.day_of_week_of_year(t)
      Timer.date_by_week(t)[2]
    end

    # [YYYY, (1..12), (1..31)]
    def self.date_by_month(t)
      [t.year, t.month, t.strftime('%-d').to_i]
    end

    def self.date_by_month_to_epoch(tuple)
      Timer.to_epoch(Time.strptime("#{tuple[0]} #{tuple[1]} #{tuple[2]}", '%Y %m %d'))
    end

    def self.month(t)
      Timer.date_by_month(t)[1]
    end

    def self.day_of_month(t)
      Timer.date_by_month(t)[2]
    end

    # (+, -)HH:MM
    def self.tz_of(t)
      t.strftime('%:z')
    end

  end
end
